  /************************************************************************
 ***************************************************************************
*****************************************************************************
*****					 				*****
*****		    Preconditioned Conjugate Gradient			*****
*****		    Double Precision (double) version.			*****
*****					 				*****
*****************************************************************************
 ***************************************************************************
  ************************************************************************/
#include "math.h"
#include "float.h"
#include "types.h"
#include "funcions.h"
static char DCGsid[] = "@(#)dcg.c	1.2  5/19/86";


int dcg( n, atimes, msolve, x, b, r, z, p, itmax, eps, err )
int (*atimes)(), (*msolve)();
int	n, itmax;
double	*x, *b, *r, *z, *p, eps, *err;
/*
  Purpose:
    This routine does preconditioned conjugate gradient iteration
    on the symmetric positive definte system Ax = b.

  Input:
    n		Size of system to be iterated (ie A is nxn).
    x		Initial guess of the solution.  If no information on the 
	        solution is available then use a zero vector or b.
    b		Right hand side.
    r           Temporary space (double r(n))
    z           Temporary space (double z(n))
    p           Temporary space (double p(n))
    eps		Requested error tollerence.  System is iterated until
		||b-Ax||/||b|| < eps.  Normal choice is 1.0e-6.
    itmax	Maximum number of iterations the user is willing to allow.

  Output:
    dcg		Returns number of iterations taken.  Error Returns follow:
    		dcg = itmax 	May have not converged.  Check err.
		dcg = -1	System size too small (n<2).
		dcg = -2	Could not get memory for tempory vectors.
		dcg = -3	Error tollerence requested too tight.
		dcg = -4	Divergence of iteration detected.
		dcg = -10	User declared error in atimes.
		dcg = -11	User declared error in msolve.
    err		L2 norm of the relative residual error estimate 
	        ||b-Ax||/||b||.  This estimate of the true error is not always
		very accurate.  Note that err is a return value so the user
		must call dcg with a pointer to err.

  User Defined Functions and Structures.
    int atimes( n, x, y )  
	        The user should write a routine that calculates y = Ax, given
		an x vector.  Return non-zero if an error occures.
    int msolve( n, z, r, atimes )
	        The user should write a routine that solves Mz=r for z, given
		a vector r.  Return non-zero if an error occures.
*/
{
  register int i;
  int j;
  double raux,bb_r,b_rnorm;
  double bnorm, bknum, bkden, bk;
  double akden, ak;
  double toobig;	/* When err est gets this big we signal divergence! */

  /* Check input data. */
  /*  if( n<2 ) return( -1 ); */
  if( eps < 1.0e-12 ) return( -3 );

  /* Calculate norm of right hand size. */
  bnorm = sqrt( ddot( n, b, 1, b, 1 ) );
  
  /* Calculate r = b - Ax and initial error. */
  if( atimes( n, x, r ) ) return( -10 );
  dexopy( n, r, b, r, 2 );
  *err = sqrt( ddot(n, r, 1, r, 1 ) )/bnorm;
  if( *err < eps ) return( 0 );
  toobig = *err * 1.0e2;

  /* Iterate!!! */
  for( i=0; i<itmax; i++ ) {

    /* Solve Mz = r. */
    if( msolve( n, z, r, atimes ) ) return( -11 );

    /* Calculate bknum = (z, Mz) and p = z (first iteration). */
    if( i == 0 ) {
      dcopy( n, z, 1, p, 1 );
      bknum = bkden = ddot( n, z, 1, r, 1 );
    } 
    else {
      /* Calculate bknum = (z, r), bkden and bk. */
      bknum = ddot( n, z, 1, r, 1 );
      bk    = bknum/bkden;
      bkden = bknum;

      /* Calculate p = z + bk*p, z = Ap, akden = (p, Ap) and ak. */
      daxpyx( n, bk, p, 1, z, 1 );
    }
    if( atimes( n, p, z ) ) return( -10 );
    akden = ddot( n, p, 1, z, 1 );
    ak    = bknum/akden;

    /* Update x and r. Calculate error. */
    daxpy( n,  ak, p, 1, x, 1 );
    daxpy( n, -ak, z, 1, r, 1 );
    
    /*    *err = sqrt( ddot( n, r, 1, r, 1 ) )/bnorm;    */
    
    /* calculem la condicio d'acabament proposada per Mauricio Resende */
    bb_r= 0.0;
    b_rnorm= 0.0;
    for(j=0;j<n;j++) {
      raux= b[j]-r[j];
      b_rnorm += raux*raux;
      bb_r += b[j]*raux;
    }
    *err= fabs(1.0 - fabs(bb_r)/(bnorm*sqrt(b_rnorm)));
    

    if( *err < eps ) return( i+1 );
    if( *err > toobig ) return( -4 );
  }				/* end for loop. */
  return( itmax );
}				/* end dcg. */
  /***************************************************************
  *****************************************************************
 *******************************************************************
*****								*****
*****				BLAS				*****
*****		Basic Linear Algebra Subroutines		*****
*****	      Written in the C Programming Language.		*****
*****		 << DOUBLE PRECISION VERSION >>			*****
*****								*****
*****	Functions include:					*****
*****	idamax, daxpy, daxpyx, dcopy, ddot, dnrm2,		*****
*****	dexopy, dfill						*****
*****								*****
 *******************************************************************
  *****************************************************************
   ***************************************************************/
#ifndef SMALLdp
#define SMALLdp		2.22507385850720e-308		/* Smallest (pos) binary float */
#endif
#ifndef HUGEdp
#define HUGEdp		1.79769313486232e+308		/* Largest binary float */
#endif

int idamax( n, sx, incx )
double	*sx;
int	n, incx;
/*
    PURPOSE
        Finds the index of element having max. absolute value.

    INPUT
	n	Number of elements to check.
	sx	Vector to be checked.
	incx	Every incx-th element is checked.

*/
{
  double smax = 0.0e0;
  int	 i, istmp = 0;

#ifndef abs
#define abs(x) ((x)>0.0?(x):-(x)) 
#endif
  if( n <= 1 ) return( istmp );
  if( incx != 1 ) {
    /* Code for increment not equal to 1. */
    if( incx < 0 ) sx = sx + ((-n+1)*incx + 1);
    istmp = 0;
    smax  = abs( *sx );
    sx += incx;
    for( i=1; i<n; i++, sx+=incx ) 
      if( abs( *sx ) > smax ) {
	istmp = i;
	smax  = abs( *sx );
      }
    return( istmp );
  }
  /* Code for increment equal to 1. */
  istmp = 0;
  smax  = abs(*sx);
  sx++;
  for( i=1; i<n; i++, sx++ ) 
    if( abs( *sx ) > smax ) { 
      istmp = i;
      smax  = abs( *sx );
    }
  return( istmp );
}
 
void daxpy( n, sa, sx, incx, sy, incy )
double *sx, *sy, sa;
int    n, incx, incy;
/*
  PURPOSE
    Vector times a scalar plus a vector.  sy = sy + sa*sx.

  INPUT
    n		Number of elements to multiply.
    sa		Scalar to multiply by.
    sx		Pointer to double vector to scale.
    incx	Storage incrament for sx.
    sy		Pointer to double vector to add.
    incy	Storage incrament for sy.

  OUTPUT
    sy		sy = sy + sa*sx
*/
{
  register int i;

  if( n<=0 || sa==0.0 ) return;
  if( incx == incy ) {
    if( incx == 1 ) {
      /* Both increments = 1 */
      for( i=0; i<n; i++,sy++,sx++ )
	*sy += sa*(*sx);
      return;
    }
    if( incx>0 ) {
      /* Equal, positive, non-unit increments. */
      for( i=0; i<n; i++,sx+=incx,sy+=incx )
	*sy += sa*(*sx);
      return;
    }
  }
  /* Unequal or negative increments. */
  if( incx < 0 ) sx += ((-n+1)*incx + 1);
  if( incy < 0 ) sy += ((-n+1)*incy + 1);
  for( i=0; i<n; i++,sx+=incx,sy+=incy ) 
    *sy += sa*(*sx);
}
 
void daxpyx( n, sa, sx, incx, sy, incy )
double *sx, *sy, sa;
int    n, incx, incy;
/*
  PURPOSE
    Vector times a scalar plus a vector.  sx = sy + sa*sx.

  INPUT
    n		Number of elements to multiply.
    sa		Scalar to multiply by.
    sx		Pointer to double vector to scale.
    incx	Storage incrament for sx.
    sy		Pointer to double vector to add.
    incy	Storage incrament for sy.

  OUTPUT
    sx		sx = sy + sa*sx
*/
{
  register int i;

  if( n<=0 || sa==0.0 ) return;
  if( incx == incy ) {
    if( incx == 1 ) {
      /* Both increments = 1 */
      for( i=0; i<n; i++, sx++, sy++ )
	*sx = *sy + sa*(*sx);
      return;
    }
    if( incx>0 ) {
      /* Equal, positive, non-unit increments. */
      for( i=0; i<n; i++, sx+=incx, sy+=incx)
	*sx = *sy + sa*(*sx);
      return;
    }
  }
  /* Unequal or negative increments. */
  if( incx < 0 ) sx += ((-n+1)*incx + 1);
  if( incy < 0 ) sy += ((-n+1)*incy + 1);
  for( i=0; i<n; i++,sx+=incx,sy+=incy ) 
    *sx = *sy + sa*(*sx);
}

void dcopy( n, sx, incx, sy, incy )
double  *sx, *sy;
int     n, incx, incy;
/*
    PURPOSE
        Copies vector sx into vector sy.
 
    INPUT
        n    Number of components to copy.
	sx   Source vector
	incx Index increment for sx.
        incy Index increment for sy.
 
    OUTPUT
        sy   Destination vector.
*/
{
  register int i;

  if( n<1  ) return;
  if( incx == incy ) {
    if( incx == 1 ) {
      /* Both increments = 1 */
      for( i=0; i<n; i++ )
	*(sy++) = *(sx++);
      return;
    }
    if( incx > 0 ) {
      /* Equal, positive, non-unit increments. */
      for( i=0; i<n; i++, sx+=incx, sy+=incx)
	*sy = *sx;
      return;
    }
  }
  /* Non-equal or negative increments. */
  if( incx < 0 ) sx += ((-n+1)*incx + 1);
  if( incy < 0 ) sy += ((-n+1)*incy + 1);
  for( i=0; i<n; i++,sx+=incx,sy+=incy ) 
    (*sx) = (*sy);
  return;
}
double ddot( n, sx, incx, sy, incy )
double	*sx, *sy;
int	n, incx, incy;
/*
    PURPOSE
        Forms the dot product of a vector.

    INPUT
        n       Number of elements to sum.
        sx      Address of first element of x vector.
        incx    Incrament for the x vector.
        sy      Address of first element of y vector.
        incy    incrament for the y vector.

    OUPUT
        sdot    Dot product x and y.  Double returned
		due to `C' language features.
*/
{
  register int i;
  double stemp = 0.0e0;

  if( n<1 ) return( stemp );
  if( incx == incy ) {
    if( incx == 1 ) {
      /* Both increments = 1 */
      for( i=0; i<n; i++, sx++, sy++ )
	stemp += (*sx)*(*sy);
      return( stemp );
    }
    if( incx>0 ) {
      /* Equal, positive, non-unit increments. */
      for( i=0; i<n; i++, sx+=incx, sy+=incx)
	stemp += (*sx)*(*sy);
      return( stemp );
    }
  }
  /* Unequal or negative increments. */
  if( incx < 0 ) sx += ((-n+1)*incx + 1);
  if( incy < 0 ) sy += ((-n+1)*incy + 1);
  for( i=0; i<n; i++,sx+=incx,sy+=incy ) 
    stemp += (*sx)*(*sy);
  return( stemp );
}				/* End of ---DDOT--- */

double dnrm2( n, sx, incx )
double	*sx;
int	n, incx;
/*
    PURPOSE
        Computes the Euclidean norm of sx while being
	very careful of distructive underflow and overflow.

    INPUT
        n       Number of elements to use.
        sx      Address of first element of x vector.
        incx    Incrament for the x vector (>0).

    OUPUT
        snrm2   Euclidean norm of sx.  Returns double
		due to `C' language features.
    REMARKS
        This algorithm proceeds in four steps.
	1) scan zero components.
	2) do phase 2 when component is near underflow.
*/
{
  register int i;
  int	 phase = 3;
  double sum = 0.0e0, cutlo, cuthi, hitst, r2mach();
  double xmax;

  if( n<1 || incx<1 ) return( sum );

  cutlo = sqrt( DBL_MIN/r2mach() );			/* Calculate near underflow */
  cuthi = sqrt( DBL_MAX );				/* Calculate near  overflow */
  hitst = cuthi/(double) n;
  i     = 0;

  /* Zero Sum. */
  while( *sx == 0.0 && i<n ) {
    i++;
    sx += incx;
  }
  if( i>=n ) return( sum );

START:
  if( abs( *sx ) > cutlo ) {
    for( ; i<n; i++, sx+=incx ) {		/* Loop over elements. */
      if( abs( *sx ) > hitst ) goto GOT_LARGE;  
      sum += (*sx) * (*sx);
    }
    sum = sqrt( sum );
    return( sum );				/* Sum completed normaly. */
  }
  else {					/* Small sum prepare for phase 2. */
    xmax  = abs( *sx );
    sx += incx;
    i++;
    sum += 1.0;
    for( ; i<n; i++, sx+=incx ) {
      if( abs( *sx ) > cutlo ) {		/* Got normal elem.  Rescale and process. */
	sum = (sum*xmax)*xmax;
	goto START;
      }
      if( abs( *sx ) > xmax ) {
	sum  = 1.0 + sum*(xmax /(*sx))*(xmax /(*sx));
	xmax = abs( *sx );
	continue;
      }
      sum += ((*sx)/xmax)*((*sx)/xmax);
    }
    return( (double)xmax*sqrt( sum ) );
  }						/* End of small sum. */

 GOT_LARGE:
  sum  = 1.0 + (sum/(*sx))/(*sx);		/* Rescale and process. */
  xmax = abs( *sx );
  sx   += incx;
  i++;
  for( ; i<n; i++, sx+=incx ) {
    if( abs( *sx ) > xmax ) {
      sum  = 1.0 + sum*(xmax /(*sx))*(xmax /(*sx));
      xmax = abs( *sx );
      continue;
    }
    sum += ((*sx)/xmax)*((*sx)/xmax);
  }
  return( (double)xmax*sqrt( sum ) );		/* End of small sum. */
}						/* End of ---DNRM2--- */

void dexopy( n, v, x, y, itype )
int	n, itype;
double	*v, *x, *y;
/*
  Purpose: v = x op y
    To operate on the vectors x and y.

  Input:
    n		Number of elements to scale.
    x		First operand vector.
    y		Second operand vector.
    itype	Type of operation to perform:
		itype = 1 => '+'
		itype = 2 => '-'

  Output:
    v		Result vector of x op y.
*/
{
  register int i;

  if( n<1 ) return;

  if( itype == 1 )	/* ADDITION. */
    for( i=0; i<n; i++, v++, x++, y++ )
      *v = *x + *y;
  else			/* SUBTRACTION. */
    for( i=0; i<n; i++, v++, x++, y++ )
      *v = *x - *y;
}

void dfill( n, v, val )
int	n;
double	*v, val;
/*
  Purpose
    To fill the DOUBLE vector v with the DOUBLE value val.
*/
{
  register int i;

  if( n<1 ) return;
  for( i=0; i<n; i++, v++ )
    *v = val;
}

double r2mach()
/* ---------------------------------------------------------------------
        This routine computes the unit roundoff for double precision 
	of the machine.  This is defined as the smallest positive 
	machine number u such that  1.0 + u .ne. 1.0
	Returns a double due to `C' language features.
----------------------------------------------------------------------*/
{
    double u = 1.0e0, comp;
 
    do {
        u *= 0.5e0;
        comp = 1.0e0 + u;
    }
    while( comp != 1.0e0 );
    return( (double)u*2.0e0 );
}
/*-------------------- end of function r2mach ------------------------*/
 



