/*****************************************************************************
  Routine for solving the multicommodity network flow problem:

        min     c'1 x1 + ...+ c'K xK 

	subj. a 

		A x1 = r1      
		  .                      Network equations for
		  .                      each commodity
		  .
		A xK = rK

		X1 + ...+ XK <= M        Linking constraints

	      X1 <= UX1, ... ,X1 <= UXk  Arc capacities

  The routine implements a specialized IP primal-dual algorithm,
  combining Cholesky factorizations and Preconditioned Conjugate
  Gradient solvers.

  The parameters are:

  nar_: number of arcs. 
  nat_: number of commodities.
  nnu_: number of nodes.
  ncb_: number of side constraints (not yet implemented; provide a 0 as input value).
  nlb_: number of nonzeros in side constraints structure (not yet implemented).
  mnk_(nar): source node for each arc (numbered 0..nar-1).
  mnl_(nar): target node for each arc (numbered de 0..nar-1).
  cap_(nar*nat): arc capacities.
  cost_(nar*nat): arc costs.
  cmt_(nar): mutual capacity constraints.
  inj_(nnu*nat): flow supplies/demands at nodes.
  lowcb_(ncb): upper bound of side constraints (not yet implemented).
  upcb_(ncb): lower bound of side constraints (not yet implemented).
  arccb_(nlb): arc of each nonzero element in side constraints structure 
               (numbered 0..nar-1) (not yet implemented).
  artcb_(nlb): commodity of each nonzero element in side constraints structure
               (numbered 0..nat-1) (not yet implemented).
  numcb_(nlb): side constraint of each nonzero element in side constraints 
               structure (numbered 0..ncb-1) (not yet implemented).
  coefcb_(nlb):coefficient of each nonzero element in side constraints structure
               (not yet implemented). 
  flux_(nar*nat): optimal flows achieved (returned value).
  inf_: infinity value considered (e.g., send a 1.0e+32)
  fobj_: optimal objective function value achieved (returned value).
  it_: number of iterations performed (returned value).
  epscg_: initial precision required to the solutions provided by the 
          conjugate gradient method. If a 0 or a negative value is provided,
	  the default parameter EPS_CG will be used. Small values (e.g. 1.0e-6)
	  provide robust but slower executions. A 1.0e-2 value works fine in
	  most cases. If you have convergence problems, reduce this to 1e-3 
	  or 1e-4.
  cgit_: number of conjugate gradient iterations performed (returned value).
  iexit_: exit status (returned value):
          iexit= -50: Not enough memory
	  iexit= -2 : Too many iterations
	  iexit= 0  : Optimal solution found
	  iexit= 5  : Quasi-optimal solution found (the required duality gap
	              has been achieved but the primal or dual problems
		      are not still feasible).
	  (note that, up to now, the routine does not detect if a problem is
	  infeasible/unbounded).
*****************************************************************************/
#include "includes.h"
void ipnet_min(int *nar_, int *nat_, int *nnu_, int *ncb_, int *nlb_, 
	       int *mnk_, int *mnl_, double *cap_, double *cost_, 
	       double *cmt_, double *inj_, double *lowcb_, double *upcb_, 
	       int *arccb_, int *artcb_, int *numcb_, double *coefcb_, 
	       double *flux_, double *inf_, double *fobj_, int *it_, 
	       int *cgit_, double *epscg_, int *iexit_)
{

  /* carreguem variables globals */
  nar= *nar_;
  nat= *nat_;
  nnu= *nnu_;
  nnux= nnu-1;
  ncb= *ncb_;
  nlb= *nlb_;
  mnk= mnk_;
  mnl= mnl_;
  cap= cap_;
  cost= cost_;
  cmt= cmt_;
  inj= inj_;
  lowcb= lowcb_;
  u_fcb= lowcb;
  upcb= upcb_;
  rhscb= upcb;
  arccb= arccb_;
  artcb= artcb_;
  numcb= numcb_;
  coefcb= coefcb_;
  flux= flux_;
  inf= *inf_;
  if ( (*epscg_) > 0.0) 
    epscg= *epscg_;
  else
    epscg= EPS_CG;
  iexit= 0;

  /* verificacio inicial de l'estructura del problema */
  ip_verifica_problema();
  COMPROVAR_IEXIT;  

  /* allocatacio inicial de memoria */
  ip_allocatar_memoria_ini();
  COMPROVAR_IEXIT;
  
  /* deteccio de constriccions inactives de capacitat mutua. Es troben
     previament les estrucutures per tenir la xarxa ordenada per origens
     i destins -que son usades posteriorment- El nombre de constriccions
     de capacitat mutues reals s'emmagatzema a nxcmt */
  ip_get_nat_pointers();
  ip_get_ipk_ipl();
  COMPROVAR_IEXIT;
  ip_preproces_cmt();
  COMPROVAR_IEXIT;

  /* mides del problema lineal */
  ip_tracta_cb();
  ip_get_ufcb();
  if (nat > 1) {
    n= nar*nat+nxcmt+nfcb;
    m= nnux*nat+nxcmt+ncb; /* nnux==nnu-1 perque eliminem darrer nus */
  }
  else {
    n= nar*nat+nfcb;
    m= nnux*nat+ncb; /* nnux== nnu-1 perque eliminem darrer nus */
  }
  nl= nfcbl;
  nu= n-nl;
  nb1= nat*nnu; /* es considera posicio ficticia nnu */
  nb2= m-nb1+nat;

  /* inicialitzacio i obtencio d'algunes dades */  
  ip_get_epsmach(&epsmach);
  if (ncb>0) ip_get_cb_estruct(); 
  COMPROVAR_IEXIT;
  ip_get_pfa_ipfa();
  COMPROVAR_IEXIT;
  ip_simbolic_ASAt();
  COMPROVAR_IEXIT;
  ip_arreglar_numeracio_pfa_ipfa();
  ip_permuta_nusos();
  ip_get_ilnz();
  COMPROVAR_IEXIT_RESTAURA;
  
  /* comencem proces iteratiu d'optimitzacio */
  ip_allocatar_memoria_opt();
  COMPROVAR_IEXIT;
  ip_inicialitza_opt();
  COMPROVAR_IEXIT;
  ip_write_dimensions();
  for(it=0;!ip_optimality();it++){
    COMPROVAR_IEXIT_BREAK;
    if (nat > 1) ip_desactivar_cmt();
    ip_get_f();
    ip_get_mu();
    ip_get_S();
    ip_get_direccio();
    COMPROVAR_IEXIT_RESTAURA; 
    ip_write_info(); 
    ip_get_alfa();
    ip_update_newpoint();
    ip_test_newpoint(); 
  }
  /* acabem */
  /* ip_ajusta_valors(); */
  mu= (double)0.0; cgit= 0; m_pw_prec= 0;
  ip_write_info(); 
  ip_write_iexit();
  ip_restaura_dades();
  ip_free_memoria();
  *it_= it;
  *cgit_= totcgit;
  *fobj_= fprimal;
  *iexit_= iexit;
}

/*****************************************************************************
  verificacio inicial del problema: 
  i)   cap arc amb capacitat no nul.la (ni individual ni mutua)
  ii)  capacitats infinites individuals es posen a MAX(totalflow,maxcap)
       (maxcap cal per si totalflow=0 en problemes amb xarxes amb cicles
       i costos negatius. Aixo podria fallar, pero, si tenim arcs infinits
       en cicles i/o amb costos negatius. L'usuari hauria d'evitar tenir
       capacitats infinites en aquest tipus de models!
  iii) comprovacio de topologia del problema correcta
*****************************************************************************/
void ip_verifica_problema()
{
  double maxcap, subst_inf;
  int i;
  
  /* comprovacio de capacitats no nules */
  maxcap= 0.0;
  for (i=0;i<nar*nat;i++){
    if (cap[i]<=0.0) {
      iexit= -4;
      return;
    }
    if (cap[i]>maxcap && cap[i]!=inf) maxcap= cap[i];
  }
  for (i=0;i<nar;i++){ 
    if (cmt[i]<=0.0) {
      iexit= -4;
      return;
    }
    if (cmt[i]>maxcap && cmt[i]!=inf) maxcap= cmt[i];
  }
  /* comprovacio de capacitats individuals no infinites; si es troba alguna
     es substitueix per MAX(totalflow,maxcap) */
  totalflow= 0.0;
  for(i=0;i<nnu*nat;i++) {
    if (inj[i]>0.0) totalflow += inj[i];
  }
  subst_inf= MAX(totalflow, maxcap);
  for (i=0;i<nar*nat;i++) 
    if (cap[i]== inf) cap[i]= subst_inf;  

  /* comprovacio topologia de la xarxa */
  for(i=0;i<nar;i++) {
    if ( mnk[i]<0 || mnk[i]>=nnu || mnl[i]<0 || mnl[i]>=nnu) {
      iexit= -5;
      return;
    }
  }

  /* no es tenen en compte side constraints de moment */
}


/*****************************************************************************
  fa un preproces per observar si hi ha constriccions de capacitat mutua
  que son per naturalesa inactives. El nombre final de constr. de cap. mutua
  reals s'emmagatzema a nxcmt, i a iactcmt[i] (i=0..nxcmt-1) s'anota a quin 
  arc correspon la i-essima constriccio de cmt activa.
*****************************************************************************/
void ip_preproces_cmt()
{    
  int i,j,k,ct;
  double maxcap, maxinflow, raux1;
  double *maxinflownode=NULL;

  TRY_ALLOC(maxinflownode,nnu,double);

  if (nat == 1) {
    nxcmt= 0;
    return;
  }
  
  /* calculem l'inflow a cada node (flux que entra a un node, tant externament
     com una injeccio, com provinent d'un altre arc) */
  for(i=0;i<nnu;i++) maxinflownode[i]= 0.0;
  for(i=0;i<nnu;i++) {
    for(k=0;k<nat;k++) if ((pinj[k])[i]>0.0) maxinflownode[i]+= (pinj[k])[i];
    for(j=inil[i];j<inil[i+1];j++) {
      raux1= 0.0;
      for(k=0;k<nat;k++) raux1 += (pcap[k])[arcl_k[j]]; 
      maxinflownode[i] += MIN(raux1,cmt[arcl_k[j]]); 
    }
  }

  /* comprovem per a cada arc si la seva cmt es activa */
  /* es fan tres simples comprovacions per a cada arc i: 
     ULL : La comprovacio 1 no pot aplicar-se perque pot haver problemes
           on el totalflow sigui 0 (problemes de recirculacio de fluxos
	   en xarxes amb cicles!)
     NO 1) si totalflow < cmt[i] 
     2) si la suma de capacitats per article de l'arc i < cmt[i]
     3) si el flux maxim que li pot arribar dels arcs que finalitzen al node 
     origen de l'arc i + les injeccions en aquest mateix node < cmt[i] */
  ct= 0;
  for(i=0;i<nar;i++){
    /* 1 */
    /*    if (totalflow <= cmt[i]) goto NEXT;*/
    /* 2 */
    maxcap= 0.0;
    for(k=0;k<nat;k++) maxcap += (pcap[k])[i]; 
    if (maxcap <= cmt[i]) goto NEXT;
    /* 3 */
    maxinflow= maxinflownode[mnk[i]];
    if (maxinflow <= cmt[i]) goto NEXT;
    
    /* si passem per aqui no hem pogut eliminar la constriccio de cmt */
    iactcmt[ct]= i;
    ct++;
  NEXT:;
  }
  nxcmt= ct;
  
  FREE(maxinflownode);
}

/*****************************************************************************
  transformacio de side constraints per tenir sempre fita superior (potser
  tambe inferior).
*****************************************************************************/
void ip_tracta_cb()
{    
  int i;

  for(i=0;i<nlb;i++) if (upcb[numcb[i]] == inf) coefcb[i]= -coefcb[i];
  for(i=0;i<ncb;i++) 
    if (upcb[i] == inf) {
      upcb[i]= -lowcb[i];
      lowcb[i]= -inf;
    }
}

/*****************************************************************************
  obtencio de u_fcb (limit superior de folgues de side constraints)
  compte: u_fcb = lowcb
*****************************************************************************/
void ip_get_ufcb()
{    
  int i;
  
  nfcb= 0;
  nfcbl= 0;
  for(i=0;i<ncb;i++) {
    if (lowcb[i] != upcb[i]) nfcb++;
    if (lowcb[i] == -inf) {
      u_fcb[i]= inf; 
      nfcbl++;
    }
    else
      u_fcb[i]= upcb[i]-lowcb[i];
  }
  nfcbu= nfcb-nfcbl;
}

/*****************************************************************************
  Rutina per allocatar memoria dels vectors que s'usaran durant tota
  l'execucio del programa.
*****************************************************************************/
void ip_allocatar_memoria_ini()
{
  
  /* allocatem vectors inicials */

  iexit= -50; /* no hi ha prou memoria */
  if (!(pfa= ALLOC(nnu,int)));      /* mantenim el darrer nus per comoditat */
  else if (!(ipfa= ALLOC(nnu,int)));/* mantenim el darrer nus per comoditat */
  else if (!(workspak= ALLOC(4*nnux,int)));
  else if (!(inik= ALLOC(nnu+1,int)));
  else if (!(mnl2= ALLOC(nar,int)));
  else if (!(arck_l= ALLOC(nar,int)));
  else if (!(inil= ALLOC(nnu+1,int)));
  else if (!(mnk2= ALLOC(nar,int)));
  else if (!(arcl_k= ALLOC(nar,int)));
  else if (!(xlnz= ALLOC(nnux+1,int)));
  else if (!(xnzsub= ALLOC(nnux+1,int)));
  else if (!(pcap= ALLOC(nat,double *)));
  else if (!(pcost= ALLOC(nat,double *)));
  else if (!(pinj= ALLOC(nat,double *)));
  else if (!(pflux= ALLOC(nat,double *)));
  else if (!(iactcmt= ALLOC(nar,int)));
  else if (!(removedcmt= ALLOC(nar,char)));
  else {
    /* fem algunes inicialitzacions sobre punters i dimensions */
    w1spak= workspak;
    w2spak= &workspak[nnux];
    w3spak= &workspak[2*nnux];
    w4spak= &workspak[3*nnux];

    iexit= 0;
  }
}

/*****************************************************************************
  Rutina per alliberar la memoria dels vectors allocatats previament
*****************************************************************************/
void ip_free_memoria()
{
  FREE(pfa);
  FREE(ipfa);
  FREE(workspak);
  FREE(inik);
  FREE(mnl2);
  FREE(arck_l);
  FREE(inil);
  FREE(mnk2);
  FREE(arcl_k);
  FREE(xlnz);
  FREE(lnzdiag);
  FREE(plnz);
  FREE(pdiag);
  FREE(xnzsub);
  FREE(nzsub);
  FREE(ajcncy);
  FREE(ajcncy2);
  FREE(pcap);
  FREE(pcost);
  FREE(pinj);
  FREE(pflux);
  FREE(iactcmt);
  FREE(removedcmt);
  FREE(ipk);
  FREE(ipl);
  FREE(inicb);
  FREE(pcb);
  FREE(ipcb);
  FREE(arcb2);
  FREE(coefcb2);
  FREE(xcmt);
  FREE(xcb);
  FREE(psk);
  FREE(s);
  FREE(z);
  FREE(w);
  FREE(f);
  FREE(y);
  FREE(r);
  FREE(dy);
  FREE(bdy);
  FREE(dx);
  FREE(dz);
  FREE(dw);
  FREE(rcg);
  FREE(zcg);
  FREE(pcg);
  FREE(d_gc);
}

/*****************************************************************************
  Rutina per escriure missatge d'exit
*****************************************************************************/
void ip_write_iexit()
{
  switch (iexit){
    /* error conditions */
  case -50:
    fprintf(stderr,"Exit: Not enough memory\n");
    break;
  case -1:
    fprintf(stderr,"Exit: Numerical inestabilities found by the PCG solver\n");
    break;
  case -2:
    fprintf(stderr,"Exit: Too many iterations\n");
    break;
  case -3:
    fprintf(stderr,"Exit: Numerical inestabilities found initializing starting point\n");
    break;
  case -4:
    fprintf(stderr,"Exit: Zero arc capacities not allowed\n");    break;
  case -5:
    fprintf(stderr,"Exit: Wrong node in network structure\n");    break;
    
    /* exit conditions */
   case 0:
    printf("Exit: optimal solution found\n");
    printf("Objective Value= %1.11e\n",fprimal);      
    printf("Conjugate Gradient Iterations= %d\n",totcgit);
    break;
  case 1:
    printf("Exit: primal infeasible - dual unbounded\n");
    break;
  case 2:
    printf("Exit: primal unbounded - dual infeasible\n");
    break;
  case 3:
    printf("Exit: primal or dual infeasible\n");
    break;
  case 4:
    printf("Exit: Suboptimal solution found\n");
    printf("Objective Value= %1.11e\n",fprimal);      
    printf("Conjugate Gradient Iterations= %d\n",totcgit);
    break;
  case 5:
    printf("Exit: Quasi-optimal solution found\n");
    printf("Objective Value= %1.11e\n",fprimal);      
    printf("Conjugate Gradient Iterations= %d\n",totcgit);
    break;
  }
}

/*****************************************************************************
  Rutina per obtenir epsilon de la maquina
*****************************************************************************/
void ip_get_epsmach(double *eps)
{
  for((*eps)=(double)1.0;((double)1.0+(*eps) > (double)1.0);(*eps) /= (double)2.0);
  (*eps) *= (double)2.0;
}

/*****************************************************************************
  Rutina per fer l'estructura de constriccions a banda
*****************************************************************************/
void ip_get_cb_estruct()
{
  int i,j,k,ii,iaux,*numcb2,ini,ifi,ct;
  double raux;

  TRY_ALLOC(inicb,ncb+1,int);
  TRY_ALLOC(arcb2,nlb,int);
  TRY_ALLOC(coefcb2,nlb,double);
  TRY_ALLOC(pcb,ncb,int);
  TRY_ALLOC(ipcb,ncb,int);
  TRY_ALLOC(numcb2,nlb,int);
  
  /* creacio de pcb i ipcb: permutacio de constriccions a banda i la seva
     permutacio inversa. Deixem primer les cb. amb folgues amb fita superior,
     despres les cb amb folgues nomes amb fita inferior, i finalment les
     cb sense folga ---cb's d'igualtat */
  for(i=0;i<ncb;i++) pcb[i]= i;
  /* primera passada: les cb's d'igualtat al final */
  if ((nfcb != ncb) && (nfcb != 0)) {
    for(i=ncb-1,ii=-1;(i>=0)&&(ii<i);i--) {
      if (fabs(u_fcb[i]) > epsmach) {
	for(ii++;(fabs(u_fcb[ii]) > epsmach)&&(ii<i);ii++);
	iaux= pcb[i];
	pcb[i]= pcb[ii];
	pcb[ii]= iaux;
	raux= u_fcb[i];
	u_fcb[i]= u_fcb[ii];
	u_fcb[ii]= raux;
	raux= rhscb[i];
	rhscb[i]= rhscb[ii];
	rhscb[ii]= raux;
      }
    }
  }
  /* segona passada: les cb's amb folgues sense fita superior al final
     de les cb's de desigualtat */
  if ((nfcbl != nfcb) && (nfcbl != 0)) {
    for(i=nfcb-1,ii=-1;(i>=0)&&(ii<i);i--) {
      if (u_fcb[i] != inf) {
	for(ii++;(u_fcb[ii] != inf)&&(ii<i);ii++);
	iaux= pcb[i];
	pcb[i]= pcb[ii];
	pcb[ii]= iaux;
	raux= u_fcb[i];
	u_fcb[i]= u_fcb[ii];
	u_fcb[ii]= raux;
	raux= rhscb[i];
	rhscb[i]= rhscb[ii];
	rhscb[ii]= raux;
      }
    }
  }
  /* creem ipcb */
  for(i=0;i<ncb;i++) ipcb[pcb[i]]= i;
  
  /* creacio de estructura de cb's, ordenades per files, amb el punter a 
     inici per cb inicb(ncb+1) */
  /* copiem vectors, amb numcb en nova ordenacio i ordre d'arcs global */
  for(i=0;i<nlb;i++) {
    numcb2[i]= ipcb[numcb[i]];
    coefcb2[i]= coefcb[i];
    arcb2[i]= nar*artcb[i]+arccb[i];
  }
  /* trobem inicb */
  for(i=0;i<ncb+1;i++) inicb[i]= 0;
  for(i=0;i<nlb;i++) inicb[numcb2[i]]++;
  for(i=0,j=0;i<ncb;i++) {
    k= inicb[i];
    inicb[i]= j;
    j += k;
  }
  inicb[ncb]= j;
  /* trobem coefcb2 i arcb2 */
  for(i=0;i<ncb;i++) {
    ini= inicb[i]; ifi= inicb[i+1];
    ct= ini;
    for(j= ini;(j<nlb)&&(ct<ifi);j++){ 
      if (numcb2[j] == i) {
	iaux= numcb2[ct];
	numcb2[ct]= numcb2[j];
	numcb2[j]= iaux;
	iaux= arcb2[ct];
	arcb2[ct]= arcb2[j];
	arcb2[j]= iaux;
	raux= coefcb2[ct];
	coefcb2[ct]= coefcb2[j];
	coefcb2[j]= raux;
	ct++;
      }
    }
  }
  FREE(numcb2);    
}

/*****************************************************************************
  Rutina per obtenir punters per article a diversos vectors
*****************************************************************************/
void ip_get_nat_pointers()
{
  int k,ja,jn;
  
  for(k=0;k<nat;k++) {
    ja= nar*k;
    jn= nnu*k;
    pcap[k]= &(cap[ja]);
    pcost[k]= &(cost[ja]);
    pflux[k]= &(flux[ja]);
    pinj[k]= &(inj[jn]);
 }
}

/*****************************************************************************
  Rutina per trobar l'estructura de dades que s'usara posteriorment, on
  es te la xarxa ordenada per origens i per destins alhora.
*****************************************************************************/
void ip_get_ipk_ipl()
{
  int i,j,ct,l;

  /* tractament de ipk i ipl */
  TRY_ALLOC(ipk,nar,K_L_ARC);
  TRY_ALLOC(ipl,nar,K_L_ARC);
  for(i=0;i<nar;i++){
    j= mnk[i];
    ipk[i].k= j; ipl[i].k= j;
    j= mnl[i];
    ipk[i].l= j; ipl[i].l= j;
    ipk[i].arc= i; ipl[i].arc= i;
  }
  qsort((char *)ipk,nar,sizeof(K_L_ARC),ip_comp_camp1);
  for(i=0,ct=0;i<nnu;i++) {
    for(inik[i]= ct,j= ct ; (ct<nar) && (ipk[ct].k == i) ; ct++); 
    l= ct-j;
    if (l>0) qsort((char *)(&ipk[j]),l,sizeof(K_L_ARC),ip_comp_camp2);
  }  
  inik[nnu]= ct;
  
  qsort((char *)ipl,nar,sizeof(K_L_ARC),ip_comp_camp2);
  for(i=0,ct=0;i<nnu;i++) {
    for(inil[i]= ct,j= ct ; (ct<nar) && (ipl[ct].l == i) ; ct++); 
    l= ct-j;
    if (l>0) qsort((char *)(&ipl[j]),l,sizeof(K_L_ARC),ip_comp_camp1);
  }  
  inil[nnu]= ct;

  for(i=0;i<nar;i++) {
    mnl2[i]= ipk[i].l;
    arck_l[i]= ipk[i].arc;
    mnk2[i]= ipl[i].k;
    arcl_k[i]= ipl[i].arc;
  }
  FREE(ipk);
  FREE(ipl);
}

/*****************************************************************************
  Rutina per trobar vector de permutacio dels nusos.
  Per tal de fer aixo considera l'algorisme del minimum degree ordering 
  aplicat al graf creat per la matriu simetrica A*A(t) (A es la matriu de
  xarxa sense considerar el darrer nus).
  Crida a la rutina genqmd() del paquet SPARSPAK.
  S'obtenen pfa i ipfa (permutacio directa i inversa):
  pfa[i]= k indica quina fila original hi ha a la posicio i (fila k)
  ipfa[k]= i indica la posicio que ocupa la fila original k ara (posicio i)
  Guardem els dos perque la permutacio no te perque ser simetrica.

  ATENCIO: donat que la rutina genqmd() esta en fortran, la numeracio de les
  permutacions dins pfa i ipfa comencen a partir de la posicio 1 i no 0.
  S'usaran amb numeracio Fortran per fer la factoritzacio simbolica amb 
  SPARSPAK, i despres ja es podran numerar segons C.
*****************************************************************************/
void ip_get_pfa_ipfa()
{    
  int i,ct,ik,fk,il,fl,nodk,nodl,nodprev,nofsub,*tmp,maxajc;
  
  /* creacio de xadj i ajcncy */
  TRY_ALLOC(xadj,nnu,int);
  TRY_ALLOC(ajcncy,2*nar,int);
  
  ct= 0;
  xadj[0]= 1;
  for(i=0;i<nnux;i++) {
    xadj[i+1]= xadj[i];
    /* fem un merge sense repeticio dels nusos de inik[i] i de inil[i] */
    ik= inik[i]; 
    fk= inik[i+1]-1;
    il= inil[i];
    fl= inil[i+1]-1;
    nodprev= 0;
    nodk= 0;
    nodl= 0;
    while ((ik <= fk) && (il <= fl) && (nodk < nnu) && (nodl <nnu)) {
      nodk= mnl2[ik]+1;
      nodl= mnk2[il]+1; 
      if (nodk < nodl) {
	if (nodk != nodprev) {
	  nodprev= nodk;
	  ajcncy[ct++]= nodk;
	  xadj[i+1]++;
	}
	ik++;
      }
      else if (nodk > nodl) {
	if (nodl != nodprev) {
	  nodprev= nodl;
	  ajcncy[ct++]= nodl;
	  xadj[i+1]++;
	}
	il++;
      }
      else /* nodk = nodl */ {
	if ((nodk != nodprev)&&(nodk != nnu)) {
	  nodprev= nodk;
	  ajcncy[ct++]= nodk;
	  xadj[i+1]++;
	}
	ik++; 
	il++;
      }
    }
    for( ; (ik <= fk) && (nodk < nnu) ; ik++) {
      nodk= mnl2[ik]+1;
      if ((nodk != nodprev)&&(nodk != nnu)) {
	nodprev= nodk;
	ajcncy[ct++]= nodk;
	xadj[i+1]++;
      }
    }
    for( ; (il <= fl) && (nodl < nnu); il++) {
      nodl= mnk2[il]+1; 
      if ((nodl != nodprev)&&(nodl != nnu)) {
	nodprev= nodl;
	ajcncy[ct++]= nodl;
	xadj[i+1]++;
      }
    }
  }
  maxajc= xadj[nnux];
  np= (maxajc-1)/2;
  TRY_REALLOC(ajcncy,maxajc,int);

  /* fem una copia de ajcncy perque genqmd() el destrueix */
  TRY_ALLOC(ajcncy2,maxajc,int);
  for(i=0;i<maxajc;i++) ajcncy2[i]= ajcncy[i];

  /* obtencio de pfa i ipfa */  
  TRY_ALLOC(tmp,2*(nnux),int);
  
  GENQMD(&nnux,xadj,ajcncy2,pfa,ipfa,tmp,&tmp[1*nnux],w1spak,w2spak,w3spak,
	  w4spak,&nofsub);
  
  FREE(tmp);
  FREE(ajcncy2);
}

/*****************************************************************************
  Rutina per fer la factoritzacio simbolica de ASA(t) (A matriu de xarxa 
  sense el darrer nus).
  Usa la rutina smbfct() del paquet SPARSPAK.
*****************************************************************************/
void ip_simbolic_ASAt()
{
  int k,flag,ind;
  double arcs,pct_acum;
  
  arcs= (double) np;
  pct_acum= PCT_FILL_IN_SUPOSAT;
  maxsub= (int)(arcs*((double)1.0+pct_acum));
  TRY_ALLOC(nzsub,maxsub,int); 
  for(flag=1;flag>0; ){
    SMBFCT(&nnux,xadj,ajcncy,pfa,ipfa,xlnz,&maxlnz,xnzsub,nzsub,&maxsub,
	    w1spak,w2spak,w3spak,&flag);
    if (flag>0) {
      FREE(nzsub);
      pct_acum= pct_acum+(double)0.10;
      maxsub= (int)(arcs*((double)1.0+pct_acum));
      TRY_ALLOC(nzsub,maxsub,int);
    }
  }
  TRY_REALLOC(nzsub,maxsub,int); 
  /* allocatem lnzdiag de forma consecutiva per tots els articles */
  TRY_ALLOC(lnzdiag,nat*(maxlnz+nnux),double);
  /* creem punters plnz(k) i pdiag(k) a lnz i diag per cada article */
  TRY_ALLOC(plnz,nat,double *);
  TRY_ALLOC(pdiag,nat,double *);
  for(k=0,ind=0;k<nat;k++) {
    plnz[k]= &lnzdiag[ind];
    ind += maxlnz;
    pdiag[k]= &lnzdiag[ind];
    ind += nnux;
  }
}

/*****************************************************************************
  Rutina per posar la numeracio de pfa i ipfa de [1:m] a [0:m-1]
*****************************************************************************/
void ip_arreglar_numeracio_pfa_ipfa()
{
  int i;
  
  for(i=0;i<nnux;i++){
    pfa[i]--;
    ipfa[i]--;
  }
  pfa[nnu-1]= nnu-1;
  ipfa[nnu-1]= nnu-1;
}

/*****************************************************************************
  Rutina per fer permutacio de nusos: permutem inj() i mnk(), mnl().
*****************************************************************************/
void ip_permuta_nusos()
{
  int i,k;
  double *dtmp, *injk;
  
  for(i=0;i<nar;i++) {
    mnk[i]= ipfa[mnk[i]];
    mnl[i]= ipfa[mnl[i]];
  }
  dtmp= (double *)workspak; 
  for(k=0;k<nat;k++) {
    injk= pinj[k];
    for(i=0;i<nnux;i++) dtmp[i]= injk[i];
    for(i=0;i<nnux;i++) injk[i]= dtmp[pfa[i]];      
  }	
}

/*****************************************************************************
  Rutina per fer despermutacio de nusos: despermutem inj() i mnk(), mnl().
*****************************************************************************/
void ip_despermuta_nusos()
{
  int i,k;
  double *dtmp, *injk;
  
  for(i=0;i<nar;i++) {
    mnk[i]= pfa[mnk[i]];
    mnl[i]= pfa[mnl[i]];
  }
  dtmp= (double *)workspak; 
  for(k=0;k<nat;k++) {
    injk= pinj[k];
    for(i=0;i<nnux;i++) dtmp[i]= injk[i];
    for(i=0;i<nnux;i++) injk[i]= dtmp[ipfa[i]];      
  }
}


/*****************************************************************************
  Rutina per obtenir el vector d'indexs a lnz() i diag(): iarc(maxiarc) i 
  ini_arc(maxlnz+1) per lnz(maxlnz); vectors iarc_incident(maxinci) 
  i ini_nus(nnux+1) per diag(nnux).
*****************************************************************************/
void ip_get_ilnz()
{
  int ct,ctini,ctfill,ctprev,i,j,l,n,ior,jor,ini,ifi,inisub,fisub;
  
  maxfillin= maxlnz-np;

  /* primer part diagonal: iarc_incident(.) ini_nus(.) */
  TRY_ALLOC(ini_nus,nnux+1,int);
  maxinci= 2*nnux;
  TRY_ALLOC(iarc_incident,maxinci,int);

  for(i=0,ct=0;i<nnux;i++){
    ior= pfa[i];
    ini_nus[i]= ct;
    for(j=inik[ior];j<inik[ior+1];j++) {
      iarc_incident[ct]= arck_l[j];
      ct++;
      if (ct >= maxinci) {
	maxinci += BLOCKSIZE;
	TRY_REALLOC(iarc_incident,maxinci,int);
      }      
    }
    for(j=inil[ior];j<inil[ior+1];j++) {
      iarc_incident[ct]= arcl_k[j];
      ct++;
      if (ct >= maxinci) {
	maxinci += BLOCKSIZE;
	TRY_REALLOC(iarc_incident,maxinci,int);
      }      
    }
  }
  ini_nus[nnux]= ct;
  maxinci= ct;
  TRY_REALLOC(iarc_incident,maxinci,int);
  
  /* despres part subdiagonal: iarc() ini_arc() */
  TRY_ALLOC(ini_arc,maxlnz+1,int);
  maxiarc= maxlnz;
  TRY_ALLOC(iarc,maxiarc,int);
  
  for(i=0,ct=0,ctini=0,ctfill=0;i<nnux;i++) {
    ior= pfa[i];
    ini= xlnz[i]-1;         /*                                       */
    ifi= xlnz[i+1]-2;	    /* xnzsub i xlnz tenen numeracio Fortran */
    inisub= xnzsub[i]-1;    /*                                       */
    fisub= inisub+ifi-ini;
    for(l=inisub;l<=fisub;l++) {
      j= nzsub[l]-1;
      jor= pfa[j];
      ini_arc[ctini]= ct;
      /* situem tots els arcs que van de (ior,jor) o de (jor,ior).
	 no cal fer cap tractament especial per no tenir en compte el nus nnu
	 ja que:
	 -ior no pot prendre mai el valor nnu, donat que iterem per i=0...nnux.
	 -jor no pot prendre mai el valor nnu, donat que prove de nzsub 
	 obtingut en la factoritzacio simbolica, i aqui no pot apareixer
	 un valor de fila de nnu --el node nnu no esta permutat per pfa */
      ctprev= ct;
      for(n=inik[ior];(n < inik[ior+1])&&(mnl2[n] <= jor);n++) {
	if (mnl2[n] == jor) {
	  iarc[ct]= arck_l[n];
	  ct++;
	  if (ct >= maxiarc) {
	    maxiarc += BLOCKSIZE;
	    TRY_REALLOC(iarc,maxiarc,int);
	  }      
	}
      }  
      for(n=inik[jor];(n < inik[jor+1])&&(mnl2[n] <= ior);n++) {
	if (mnl2[n] == ior) {
	  iarc[ct]= arck_l[n];
	  ct++;
	  if (ct >= maxiarc) {
	    maxiarc += BLOCKSIZE;
	    TRY_REALLOC(iarc,maxiarc,int);
	  }      
	}
      }  
      /* avancem comptador de nou element de lnz i mirem si era fill-in*/
      ctini++;
      if (ct == ctprev) ctfill++;
    }
  }
  ini_arc[ctini]= ct;

  FREE(inik);
  FREE(mnl2);
  FREE(arck_l);
  FREE(inil);
  FREE(mnk2);
  FREE(arcl_k);
  if (ctini != maxlnz) {
    fprintf(stderr,"Internal Error (1) at ip_get_ilnz()\n");
    exit(-1);
  }
  if (ctfill != maxfillin) {
    fprintf(stderr,"Internal Error (2) at ip_get_ilnz()\n");
    exit(-1);
  }
}

/*****************************************************************************
  Rutina per escriure linees d'informacio
*****************************************************************************/
void ip_write_dimensions()
{
  int i;
  double arops,raux;
  
  arops= m+3*maxlnz;
  for(i=0;i<nnux;i++) {
    raux= xlnz[i+1]-xlnz[i]+1;
    arops += (raux*raux);
  }

  printf("----------------------------------------------------------------------------------------------\n");
  printf("arcs= %d  nodes= %d  commodities= %d  side constraints= %d  \n",
	 nar,nnu,nat,ncb);
  printf("constraints= %-15d    variables= %d (upper= %d ; lower= %d)\n",
	 m,n,nu,nl);
  printf("mutual capacity constraints eliminated by presolve= %d \n",nar-nxcmt);
  printf("subnonz(AS(k)A')= %-13d subnonz(L(k)L(k)')= %-15d\n",np,maxlnz);
  printf("fill-in(k)= %4.1f%%               arithmetic_ops(k)= %-30.0f\n",100.0*(double)maxfillin/(double)maxlnz,arops);
  printf("density(AS(k)A')= %4.1f%%\n",100.0*(double)np/(double)(nnux*(nnux-1)/2.0));
  printf("density(L(k)L(k)')= %4.1f%%\n",100.0*(double)maxlnz/(double)(nnux*(nnux-1)/2.0));
  printf("----------------------------------------------------------------------------------------------\n");
}

/*****************************************************************************
  Rutina per escriure linees d'informacio
*****************************************************************************/
void ip_write_info()
{
  static char pf[]="PF",df[]="DF",buit[]="  ";
  char *status1,*status2,strgap[9];

  if ((it%output_freq) != 0) return;
  if ((ctline % 100) == 0) {
    ctline= 0;
    printf("Iter    mu       Primal Obj. (Infeas)     Dual Obj.  (Infeas)   Status    Gap      CGit   m_pw\n");
    printf("----------------------------------------------------------------------------------------------\n");
  }
  if (zpf) 
    status1= pf;
  else
    status1= buit;
  if (zdf) 
    status2= df;
  else
    status2= buit;
/*
  if (zpf && zdf) 
    sprintf(strgap,"%1.2e",gap);
  else
    strgap[0]= '\0';
*/
  sprintf(strgap,"%1.2e",gap);
  printf("%-4d %1.2e %14e %1.2e %14e %1.2e  %3s %3s  %8s  %5d %6d\n",it,mu,
	 fprimal,pinf,fdual,dinf,status1,status2,strgap,cgit,m_pw_prec);
  ctline++;
}

/*****************************************************************************
  Rutina per restaurar les dades d'entrada i deixar-les en l'estat original:
  despermuta els nusos, i deixa les side constraints com al principi
*****************************************************************************/
void ip_restaura_dades()
{

  ip_despermuta_nusos();
  ip_restaura_cb();

}

/*****************************************************************************
  Rutina per restaurar els limits de side constraints.
*****************************************************************************/
void ip_restaura_cb()
{
  int i;

  for(i=0;i<ncb;i++) {
    if (lowcb[i] == inf) 
      lowcb[i]= -inf;
    else
      lowcb[i]= upcb[i]-lowcb[i];
  }
}

/*****************************************************************************
  Rutina per allocatar memoria dels vectors que s'usaran durant 
  la optimitzacio
*****************************************************************************/
void ip_allocatar_memoria_opt()
{
  int i,pn,pm;

  iexit= -50; /* no hi ha prou memoria */
  if (!(s= ALLOC(n,double)));
  else if (!(z= ALLOC(n,double)));
  else if (!(w= ALLOC(nu,double)));
  else if (!(f= ALLOC(nu,double)));
  else if (!(y= ALLOC(m+nat,double)));   /* es considera posicio ficticia nnu */
  else if (!(r= ALLOC(n,double)));
  else if (!(bdy= ALLOC(m+nat,double))); /* es considera posicio ficticia nnu */
  else if (!(dy= ALLOC(m+nat,double)));      /* es considera posicio ficticia nnu */
  else if (!(dx= ALLOC(n+nat,double)));  /* es considera posicio ficticia nnu */
  /* en aquest cas cal perque dx s'usa com espai temporal de treball 
     a ip_get_direccio */
  else if (!(dz= ALLOC(n,double)));
  else if (!(dw= ALLOC(nu,double)));
  else if ((nxcmt>0)&&(!(xcmt= ALLOC(nxcmt,double))));
  else if ((nfcb>0)&&(!(xcb= ALLOC(nfcb,double)))); 
  else if (!(psk= ALLOC(nat,double *)));
  else if (!(beta= ALLOC(m+nat,double))); /* es considera posicio ficticia nnu */
  else if (!(rcg= ALLOC(nb2+1,double)));
  else if ((nb2>0)&&(!(zcg= ALLOC(nb2,double))));
  else if ((nb2>0)&&(!(pcg= ALLOC(nb2,double))));
  else if ((nb2>0)&&(!(d_gc= ALLOC(nb2,double))));
  else {
    /* inicialitzem punters per cmt i cb als vectors anteriors */
    for(i=0,pn=0;i<nat;i++,pn += nar) psk[i]= &(s[pn]);
    pm= nnu*nat;  /* es considera posicio ficticia nnu */
    pn= nar*nat;
    if (nat > 1) {
      ycmt= &(y[pm]);
      dycmt= &(dy[pm]);
      bdycmt= &(bdy[pm]);
      scmt= &(s[pn]); 
      rcmt= &(r[pn]);
      zcmt= &(z[pn]); 
      dzcmt= &(dz[pn]); 
      dxcmt= &(dx[pn]);
      wcmt= &(w[pn]);
      fcmt= &(f[pn]);
    }

    pm += nxcmt;
    pn += nxcmt;
    ycb= &(y[pm]);
    dycb= &(dy[pm]);
    bdycb= &(bdy[pm]);
    scb= &(s[pn]); 
    rcb= &(r[pn]); 
    zcb= &(z[pn]); 
    dzcb= &(dz[pn]); 
    dxcb= &(dx[pn]); 
    wcb= &(w[pn]);
    fcb= &(f[pn]);

    iexit= 0;
  }
}

/*****************************************************************************
  Rutina per inicialitzar dades per optimitzacio
*****************************************************************************/
void ip_inicialitza_opt()
{
  int i,ichol;
  double rho_x0,mu_x0,epscg_aux;

  /* inicialitzem x, solucionant el problema 
     min c'x+rho/2 (||x||+||f||)
     subj. a Ax=b
     Per aixo cal calcular (y es un vector de multiplicadors qualsevol, no 
     te res a veure amb l'y de la resta de l'algorisme)
     i)  y= (AA')^{-1}*(b/(2*rho)+Ac-Au/rho)
     ii) x= 1/2*(u+(A'y-c)/rho)  
     
     S'usen dy,bdy,dx com espai temporals */
  
  /* rho_x0 */
  rho_x0= -100;

  /* calcul y= (AA')^{-1}*(b/(2*rho)+Ac-Au/rho) 
     el terme de la dreta s'emmagatzema a bdy
     el vector solucio y es troba a dy */
  for(i=0;i<n;i++) s[i]= 1.0;
  escal_s= 1.0;
  for(i=0;i<nar*nat;i++) dx[i]= cost[i]-(cap[i]/rho_x0);
  for(i=0;i<nxcmt;i++) dxcmt[i]= -(cmt[iactcmt[i]]/rho_x0);
  ip_get_Ax(bdy,bdycmt,bdycb,dx,dxcmt,dxcb);
  for(i=0;i<nnu*nat;i++) bdy[i] += inj[i]/rho_x0;
  for(i=0;i<nxcmt;i++) bdycmt[i] += (cmt[iactcmt[i]]/rho_x0);
  epscg_aux= epscg; /* salvem valor de epscg */
  epscg= 0.0;       /* i utilitzem maxima precissio al punt inicial */
  ip_get_dy(&ichol);
  epscg= epscg_aux; /* recuperem valor original de epscg */
  if (ichol==1) {
    iexit= -3;
  }
  /* calcul  x= 1/2*(u+(A'y-c)/rho) 
     el vector y es troba a dy 
     la solucio x es deixa a flux i xcmt */
  ip_get_Aty(dx,dxcmt,dxcb,dy,dycmt,dycb);
  for(i=0;i<nar*nat;i++) flux[i]= 0.5*(cap[i]+ (dx[i]-cost[i])/rho_x0);
  for(i=0;i<nxcmt;i++) xcmt[i]= 0.5*(cmt[iactcmt[i]]+dxcmt[i]/rho_x0);

  /* fem interiors les components fora de fites */
  for(i=0;i<nar*nat;i++) {
    if ((flux[i]<= EPSIL_X0)||(flux[i]>=cap[i]-EPSIL_X0))
      flux[i]= MIN(cap[i]/2.0,1.0);
  }
  for(i=0;i<nxcmt;i++) {
    if ((xcmt[i]<= EPSIL_X0)||(xcmt[i]>=cmt[iactcmt[i]]-EPSIL_X0))
      xcmt[i]= MIN(cmt[iactcmt[i]]/2.0,1.0);
  }
    
  /* no hi ha cb de moment */

  for(i=0;i<nfcbu;i++) xcb[i]= MIN(u_fcb[i]/(double)2.0,X0);
  for(i=nfcbu;i<nfcb;i++) xcb[i]= X0;
  
  /* inicialitzem variables z,w,y: y=0, i z,w propers a central path */  
  for(i=0;i<m+nat;i++) y[i]= (double)0.0; /* es considera posicio ficticia nnu */
  mu_x0= 100.0;
  for(i=0;i<nar*nat;i++) {
    z[i]= mu_x0/cap[i]+cost[i]/2.0+sqrt(mu_x0*mu_x0/(cap[i]*cap[i])+
					cost[i]*cost[i]/4.0);
    w[i]= mu_x0*z[i]/(cap[i]*z[i]-mu_x0);
    /*    printf("%d   z[]= %f   w[]= %f   z*x[]=%f  f*w[]=%f z-w-c[]=%f   \n",i,z[i],w[i],z[i]*flux[i],(cap[i]-flux[i])*w[i], z[i]-w[i]-cost[i]);   */
  }  
  if (nat > 1) {
    for(i=0;i<nxcmt;i++) {    
      zcmt[i]= 2.0*mu_x0/cmt[iactcmt[i]];
      wcmt[i]= 2.0*mu_x0/cmt[iactcmt[i]];
      /*      printf("%d   zcmt[]= %f   wcmt[]= %f   zcmt*xcmt[]=%f  fcmt*wcmt[]=%f zcmt-wcmt[]=%f   \n",i,zcmt[i],wcmt[i],zcmt[i]*xcmt[i],(cmt[iactcmt[i]]-xcmt[i])*wcmt[i], zcmt[i]-wcmt[i]);   */
    }
  }
  /* no hi ha cb de moment */
  for(i=0;i<nfcbu;i++) {
    zcb[i]= X0;
    wcb[i]= X0;
  }
  for(i=nfcbu;i<nfcb;i++) zcb[i]= X0_FCB;

  /* inicialitzem direccio inicial dy */
  for(i=0;i<m+nat;i++) dy[i]=(double)0.0; /* es considera posicio ficticia nnu */

  /* escalars */
  sigma= SIGMA;
  rho= RHO;
  optim_gap= OPTIM_GAP;
  for(i=0,normb=(double)0.0;i<nat;i++) normb += ip_norma2(pinj[i],nnux);
  normb += ip_norma2(rhscb,ncb);
  if (nat > 1) {
    for(i=0;i<nxcmt;i++) normb += cmt[iactcmt[i]]*cmt[iactcmt[i]];
  }
  normb= (double)1.0+sqrt(normb);
  normc= (double)1.0+sqrt(ip_norma2(cost,nat*nar));
  optim_pfeas= OPTIM_PFEAS;
  optim_dfeas= OPTIM_DFEAS;
  ctline= 0;
  output_freq= OUTPUT_FREQ;
  zpf= FALSE;
  zdf= FALSE;
  pinf= DBL_MAX;
  dinf= DBL_MAX;
  totcgit= 0;
  tipusprec= 0;
  m_pw_prec= 1;
}

/*****************************************************************************
  Rutina per verificar si ja estem al punt optim. Comprova si es satisfa
  la factibilitat primal, factibilitat dual i la reduccio de gap entre
  la f.obj. primal i dual. Calcula ja els vectors b3 (carregat a bdy) i
  b4 (carregat a dz).
*****************************************************************************/
char ip_optimality()
{
  int i,k,l;
  double *b3, *b3cmt, *b3cb, *b4, *b4cmt, *b4cb, raux, *injk;

  if (it>MAXITER) {
    iexit= -2;
    return(FALSE);
  }

  b3= bdy; 
  b3cmt= bdycmt; 
  b3cb= bdycb;
  
  b4= dz;
  b4cmt= dzcmt;
  b4cb= dzcb;

  /* factibilitat primal */
  ip_get_Ax(b3,b3cmt,b3cb,flux,xcmt,xcb);
  l= 0;
  for(k=0;k<nat;k++) {
    injk= pinj[k];
    for(i=0;i<nnux;i++) {
      b3[l]= injk[i]-b3[l]; /* equacions de xarxa */
      l++;
    }
    l++; /* es considera posicio ficticia nnu */
  }
  if (nat > 1) {
    for(i=0;i<nxcmt;i++) b3cmt[i]= cmt[iactcmt[i]]-b3cmt[i]; 
                            /* constriccions de capacitat mutua */
  }
  for(i=0;i<ncb;i++) b3cb[i]= rhscb[i]-b3cb[i]; /* constriccions a banda */
  raux= sqrt(ip_norma2(b3,m+nat))/normb; /* es considera posicio ficticia nnu */
  /*
  if ( (raux-pinf)/((double)1.0+raux) > EPS_PINF) {
    if (zdf) 
      if (zpf)
	iexit= 4;
      else
	iexit= 1;
    else 
      iexit= 3;
  } 
  */
  pinf= raux;
  zpf= (pinf < optim_pfeas);
  
  /* factibilitat dual */
  ip_get_Aty(b4,b4cmt,b4cb,y,ycmt,ycb);
  for(i=0;i<nar*nat;i++) b4[i]= cost[i]-b4[i]-z[i]+w[i]; /* fluxos */
  if (nat > 1) {
    for(i=0;i<nxcmt;i++) b4cmt[i]= -b4cmt[i]-zcmt[i]+wcmt[i];
                                                  /* folgues cap. mutua */
  }
  for(i=0;i<nfcbu;i++) b4cb[i]= -b4cb[i]-zcb[i]+wcb[i];/* folgues cb (upper) */
  for(i=nfcbu;i<nfcb;i++) b4cb[i]= -b4cb[i]-zcb[i]; /* folgues cb (lower) */
  raux= sqrt(ip_norma2(b4,n))/normc;
  /*
  if ( (raux-dinf)/((double)1.0+raux) > EPS_DINF) {
    if (zpf)
      if (zdf)
	iexit= 4;
      else
	iexit= 2;
    else 
      iexit= 3;
  } 
  */
  dinf= raux;
    zdf= (dinf < optim_dfeas);

  /* comprovacio gap */
  /* calcul funcio primal */
  for(i=0,fprimal=(double)0.0;i<nar*nat;i++) fprimal += cost[i]*flux[i];  /* fluxos */

  /* calcul funcio dual */ 
  l= 0;
  fdual= (double)0.0;
  for(k=0;k<nat;k++) {
    injk= pinj[k];
    for(i=0;i<nnux;i++) {
      fdual += injk[i]*y[l];                     /* xarxa */
      l++;
    }
    l++; /* es considera posicio ficticia nnu */
  }
    
  if (nat > 1) {
    for(i=0;i<nxcmt;i++) fdual += cmt[iactcmt[i]]*ycmt[i];  /* capacitat mutua */
  }
  for(i=0;i<ncb;i++) fdual += rhscb[i]*ycb[i];   /* constriccions a banda */
  
  for(i=0; i<nar*nat;i++) fdual -= cap[i]*w[i];  /* fluxos */
  if (nat > 1) {  
    for(i=0; i<nxcmt;i++) fdual -= cmt[iactcmt[i]]*wcmt[i]; /* capacitat mutua */
  }
  for(i=0;i<nfcbu;i++) fdual -= u_fcb[i]*wcb[i]; /* constriccions a banda */
  
  gap= fabs(fprimal-fdual)/((double)1.0+fabs(fprimal)); 
  sig_dig= (int) (MAX(0,-log10(gap)));
  /* modifiquem optim_gap si el problema conte moltes restriccions cmt 
     actives i estem proper a l'optim */
  if ((gap < OPTIM_GAP_SAFEGUARD) && 
      ((nxcmt > MAXCMT) ||(nxcmt > PCT_MAXCMT*nar))) {
    optim_gap= OPTIM_GAP_SAFEGUARD;
  }
  zgap= (gap < optim_gap);

  if (zgap && !(zpf && zdf)) iexit= 5; 
  return(zpf && zdf && zgap);
}

/*****************************************************************************
  Rutina per calcular b=A*x.
*****************************************************************************/
void ip_get_Ax(double *b_net, double *b_cm, double *b_cb, double *x_arc, 
	       double *x_cm, double *x_cb)
{
  int i,j,k,pnu,pna;
  double *vb,*vx,raux;
  
  /* calcul de b_net */
  for(i=0;i<nnu*nat;i++) b_net[i]= (double)0.0; /* es considera posicio ficticia nnu */
  pnu= 0;
  pna= 0;
  for(k=0;k<nat;k++) {
    vb= &(b_net[pnu]);
    vx= &(x_arc[pna]);
    for(i=0;i<nar;i++) {
      raux= vx[i];
      vb[mnk[i]] += raux;
      vb[mnl[i]] -= raux;
    }
    vb[nnux]= (double)0.0;
    pnu += nnu;
    pna += nar;
  }

  /* calcul de b_cm */
  if (nat >1) {
    for(i=0;i<nxcmt;i++) b_cm[i]= x_arc[iactcmt[i]]+x_cm[i]; /* folgues i 1r article */
    pna= nar;
    for(k=1;k<nat;k++) {
      vx= &(x_arc[pna]);
      for(i=0;i<nxcmt;i++) b_cm[i] += vx[iactcmt[i]];  /* resta d'articles */
      pna += nar;
    }
  }

  /* calcul de b_cb */
  for(i=0;i<ncb;i++) {
    raux= (double)0.0;
    for(j= inicb[i];j<inicb[i+1];j++) raux += x_arc[arcb2[j]]*coefcb2[j];
    b_cb[i]= raux;
  }
  for(i=0;i<nfcb;i++) b_cb[i] += x_cb[i];
}

/*****************************************************************************
  Rutina per calcular c=A'*y.
*****************************************************************************/
void ip_get_Aty(double *c_arc, double *c_cm, double *c_cb, 
		double *y_net, double *y_cm, double *y_cb)
{
  int i,j,k,pna,pnu;
  double *vc,*vy;

  /* part afectada per y_net */
  pnu= 0;
  pna= 0;
  for(k=0;k<nat;k++) {
    vc= &(c_arc[pna]);
    vy= &(y_net[pnu]);
    vy[nnux]= (double)0.0;
    for(i=0;i<nar;i++) { /* c_arc */
      vc[i]= (double)0.0;
      vc[i] += vy[mnk[i]];  /* es considera posicio ficticia nnu */
      vc[i] -= vy[mnl[i]];  /* es considera posicio ficticia nnu */
    }
    pnu += nnu;
    pna += nar;
  }

  /* part afectada per y_cm */
  if (nat > 1) {
    pna= 0;
    for(k=0;k<nat;k++) {
      vc= &(c_arc[pna]);
      for(i=0;i<nxcmt;i++) vc[iactcmt[i]] += y_cm[i]; /* c_arc */
      pna += nar;
    }
    for(i=0;i<nxcmt;i++) c_cm[i]= y_cm[i];  /* c_cm */
  }

  /* part afectada per y_cb */
  for(i=0;i<ncb;i++) {
    for(j=inicb[i];j<inicb[i+1];j++) c_arc[arcb2[j]] += y_cb[i]*coefcb2[j];
                                                   /* c_arc */
  }
  for(i=0;i<nfcb;i++) c_cb[i]= y_cb[i];  /* c_cb */
 
}

/*****************************************************************************
  Rutina per verificar si podem desactivar constriccions de capacitat
  mutua que a l'optim seran inactives
*****************************************************************************/
void ip_desactivar_cmt()
{
  int i,nremoved,ct,pna,j,k;
  double raux,*tmp;

  if (gap > 1.0e-0) return;

  nremoved= 0;
  raux= 0.0;
  for(i=0;i<nxcmt;i++) removedcmt[i]= FALSE;
  for(i=0;i<nxcmt;i++) {
    if ((xcmt[i] > 0.1*cmt[iactcmt[i]]) && (fabs(ycmt[i])<1.0e-2)) {
      nremoved++;
      removedcmt[i]= TRUE;
      raux += cmt[iactcmt[i]]*cmt[iactcmt[i]];
    }    
  }
  if (nremoved > 0) {
    printf("  Removed Mutual Capacity Constraints= %d\n",nremoved);
    /* acualitzem normb */
/*    normb= 1.0+sqrt((normb-1)*(normb-1)-raux); */
    
    /* compactem tots els vector afectats per l'eliminacio de les 
       constriccion de cmt: iactcmt,xcmt,ycmt,bdycmt,dzcmt,zcmt,wcmt,dycmt 
       Tambe modifiquem el vector dz d'infactibilitats duals (s'ha
       d'eliminar l'afectacio de les variables ycmt de les restricions
       de cmt ara desactivades) */
    ct= 0;
    for(i=0;i<nxcmt;i++) {
      if (!removedcmt[i]) {
	iactcmt[i-ct]= iactcmt[i];
	xcmt[i-ct]= xcmt[i];
	ycmt[i-ct]= ycmt[i];
	zcmt[i-ct]= zcmt[i];
	wcmt[i-ct]= wcmt[i];
	bdycmt[i-ct]= bdycmt[i]; /* emmagatzema b3cmt */
	dzcmt[i-ct]= dzcmt[i]; /* emmagatzema b4cmt */
	dycmt[i-ct]= dycmt[i]; /* emmagatzema el punt anterior per PCG */
      }
      else {
	ct++;  
	/* eliminem afectacio de ycmt a dz (infactibilitats duals) */
	pna= 0;
	j= iactcmt[i];
	raux= ycmt[i];
	for(k=0;k<nat;k++) {
	  dz[pna+j] += raux;
	  pna += nar;  
	}
      }
    }
    /* arreglem dimensions del nou problema */
    nxcmt -= nremoved;
    n -= nremoved;
    m -= nremoved;
    nu= n-nl;
    nb2 = m-nb1+nat;
  }
}

/*****************************************************************************
  Rutina per calcular vector f=u-xu
*****************************************************************************/
void ip_get_f()
{
  int i;

  for(i=0;i<nar*nat;i++) f[i]= cap[i]-flux[i];
  if (nat > 1) {
    for(i=0;i<nxcmt;i++) fcmt[i]= cmt[iactcmt[i]]-xcmt[i];
  }
  for(i=0;i<nfcbu;i++) fcb[i]= u_fcb[i]-xcb[i];
}

/*****************************************************************************
  Rutina per obtenir la constant de penalitzacio mu de la barrera logarismica
*****************************************************************************/
void ip_get_mu()
{
  int i;
  double r;

  if (zpf && zdf) {
    r= fabs(fprimal-fdual);
    mu= sigma*r/(n+nu);
  }
  else {
    r= (double)0.0;
    for(i=0;i<nar*nat;i++)  r += z[i]*flux[i];
    if (nat > 1) {
      for(i=0;i<nxcmt;i++)  r += zcmt[i]*xcmt[i];
    }
    for(i=0;i<nfcb;i++)  r += zcb[i]*xcb[i];
    for(i=0;i<nu;i++) r += f[i]*w[i];
    mu= sigma*r/(n+nu);
  }
}

/*****************************************************************************
  Rutina per crear el vector S i emmagatzemar el maxim element en valor
  absolut per fer un escalat del sistema (ASA(t)) x= b (escal_s).
*****************************************************************************/
void ip_get_S()
{
  int i;
  
  escal_s= (double)0.0;

  /* part S_u */
  for(i=0;i<nar*nat;i++) {
    s[i]= f[i]*flux[i]/(z[i]*f[i]+flux[i]*w[i]);
    if (s[i] > escal_s) escal_s= s[i];
  }
  if (nat > 1) {
    for(i=0;i<nxcmt;i++) { 
      scmt[i]= fcmt[i]*xcmt[i]/(zcmt[i]*fcmt[i]+xcmt[i]*wcmt[i]);
      if (scmt[i] > escal_s) escal_s= scmt[i];
    }
  }
  for(i=0;i<nfcbu;i++) {
    scb[i]= fcb[i]*xcb[i]/(zcb[i]*fcb[i]+xcb[i]*wcb[i]);
    if (scb[i] > escal_s) escal_s= scb[i];
  }
  
  /* part S_l */
  for(i=nfcbu;i<nfcb;i++) {
    scb[i]= xcb[i]/zcb[i];
    if (scb[i] > escal_s) escal_s= scb[i];
  }

  for(i=0;i<n;i++) s[i] /= escal_s; 
}

/*****************************************************************************
  Rutina per obtenir la direccio de descens de Newton: dy,dx,dw,dz
*****************************************************************************/
void ip_get_direccio()
{

  int i,j,ichol;
  double *tmp;
  
  /* calculem terme de la dreta b3+ASr de (ASA')dy=b3+ASr  */
  tmp= dx; /* dx s'usa com espai de treball temporal */
  ip_get_Sr();
  j= nnu*nat;
  ip_get_Ax(tmp,&(tmp[j]),&(tmp[j+nxcmt]),r,rcmt,rcb);
  for(i=0;i<m+nat;i++) bdy[i] = bdy[i]/escal_s+tmp[i]; /* es considera posicio ficticia nnu */

  /* calculem dy */
  ichol= 0;
  ip_get_dy(&ichol);
  if (ichol==1) {
    iexit= -1;
    return;
  }
  
  /* dx i comencem dz */
  ip_get_Aty(dx,dxcmt,dxcb,dy,dycmt,dycb);
  for(i=0;i<n;i++){
    dz[i] -= dx[i];
    /* per aquest ordre: primer dz[i] i despres dx[i] */
    dx[i]= (s[i]*dx[i]-r[i])*escal_s;
  }
  
  /* dw i acabem dz */
  for(i=0;i<nu;i++) {
    dw[i]= (mu+w[i]*dx[i])/f[i]-w[i];
    /* per aquest ordre: primer dw[i] i despres dz[i] */
    dz[i] += dw[i];
  }
}

/*****************************************************************************
  Rutina per calcular la passa maxima alfa
*****************************************************************************/
void ip_get_alfa()
{
  int i;
  double alfapd;
  
  alfa_p= DBL_MAX;
  alfa_d= DBL_MAX;

  for(i=0;i<nar*nat;i++) {
    if (dx[i]< -EPS_DIR) 
      alfa_p= MIN(alfa_p,-flux[i]/dx[i]);
    else if (dx[i]> EPS_DIR) 
      alfa_p= MIN(alfa_p,(cap[i]-flux[i])/dx[i]);
    else
      dx[i]= (double)0.0; 
  }
  if (nat > 1) {
    for(i=0;i<nxcmt;i++) {
      if (dxcmt[i]< -EPS_DIR) 
	alfa_p= MIN(alfa_p,-xcmt[i]/dxcmt[i]);
      else if (dxcmt[i]> EPS_DIR) 
	alfa_p= MIN(alfa_p,(cmt[iactcmt[i]]-xcmt[i])/dxcmt[i]);
      else
	dxcmt[i]= (double)0.0;
    }
  }
  for(i=0;i<nfcbu;i++) {
    if (dxcb[i]< -EPS_DIR) 
      alfa_p= MIN(alfa_p,-xcb[i]/dxcb[i]);
    else if (dxcb[i]> EPS_DIR) 
      alfa_p= MIN(alfa_p,(u_fcb[i]-xcb[i])/dxcb[i]);
    else
      dxcb[i]= (double)0.0;
  }
  for(i=nfcbu;i<nfcb;i++) {
    if (dxcb[i]< -EPS_DIR) 
      alfa_p= MIN(alfa_p,-xcb[i]/dxcb[i]);
    else if (dxcb[i]< EPS_DIR)
      dxcb[i]= (double)0.0;
  }

  for(i=0;i<n;i++) {
    if (dz[i]< -EPS_DIR) 
      alfa_d= MIN(alfa_d,-z[i]/dz[i]);
    else if (dz[i]< EPS_DIR) 
      dz[i]= (double)0.0;
  }

  for(i=0;i<nu;i++) {   
    if (dw[i]< -EPS_DIR) 
      alfa_d= MIN(alfa_d,-w[i]/dw[i]);
    else if (dw[i]< EPS_DIR)
      dw[i]= (double)0.0;
  }
  
  alfa_p= MIN(alfa_p*rho,(double)1.0);
  alfa_d= MIN(alfa_d*rho,(double)1.0);
/*
  alfapd= MIN(alfa_p,alfa_d);
  sigma= (((double)1.1-alfapd)/((double)10.0*alfapd+(double)1.0));
  sigma= sigma * sigma;
*/

  sigma= SIGMA;
}

/*****************************************************************************
  Rutina per actualitzar el nou punt amb l'alfa trobada
*****************************************************************************/
void ip_update_newpoint()
{
  int i;

  for(i=0;i<nar*nat;i++){
    flux[i] += alfa_p*dx[i];
  }
  if (nat > 1) {
    for(i=0;i<nxcmt;i++){
      xcmt[i] += alfa_p*dxcmt[i];
    }
  }
  for(i=0;i<nfcb;i++){
    xcb[i] += alfa_p*dxcb[i];
  }
  for(i=0;i<n;i++) z[i] += alfa_d*dz[i];
  for(i=0;i<nu;i++) w[i] += alfa_d*dw[i];
  for(i=0;i<m+nat;i++) y[i] += alfa_d*dy[i]; /* es considera posicio ficticia nnu */
}

/*****************************************************************************
  Rutina per testejar newpoint esta entre fites
*****************************************************************************/
void ip_test_newpoint()
{
  int i;
  double eptest,pctest;
  
  eptest= epsmach;
  pctest= (double)1.0-eptest;

  for(i=0;i<nar*nat;i++){
    if (flux[i]<=(double)0.0) 
      flux[i]= eptest;
    else if (flux[i]>=cap[i])
      flux[i]= cap[i]*pctest;
  }
  if (nat > 1) {
    for(i=0;i<nxcmt;i++){
      if (xcmt[i]<=(double)0.0)
	xcmt[i]= eptest;
      else if (xcmt[i]>=cmt[iactcmt[i]])
	xcmt[i]= cmt[iactcmt[i]]*pctest;
    }
  }
  for(i=0;i<nfcbu;i++){
    if (xcb[i]<=(double)0.0) 
      xcb[i]= eptest;
    else if (xcb[i]>=u_fcb[i]) 
      xcb[i]= u_fcb[i]*pctest;
  }
  for(i=nfcbu;i<nfcb;i++){
    if (xcb[i]<=(double)0.0) 
      xcb[i]= eptest;
  }
  for(i=0;i<n;i++) {
    if (z[i]<=(double)0.0) 
      z[i]= eptest;
  }
  for(i=0;i<nu;i++) {
    if (w[i]<=(double)0.0) 
      w[i]= eptest;
  }
}

/*****************************************************************************
  Rutina per calcular el vector Sr=S(ru,rl)
*****************************************************************************/
void ip_get_Sr()
{
  int i;

  /* b4 esta a dz */
  for(i=0;i<nar*nat;i++) r[i]= s[i]*(dz[i]-mu/flux[i]+z[i]);
  if (nat > 1) {
    for(i=0;i<nxcmt;i++) rcmt[i]= scmt[i]*(dzcmt[i]-mu/xcmt[i]+zcmt[i]);
  }
  for(i=0;i<nfcb;i++) rcb[i]= scb[i]*(dzcb[i]-mu/xcb[i]+zcb[i]);

  for(i=0;i<nu;i++) r[i] += s[i]*(mu/f[i]-w[i]);
}

/*****************************************************************************
  Rutina per calcular dy: (ASA')dy = b3+ASr
  La part de la dreta b3+ASr ja es troba emmagatzemada al vector bdy,
  i en acabar a dy hi ha la solucio del sistema.
  En retornar, si iret==0 tot ha anat be. Si iret==1 hi ha hagut problemes
  !!encara no hi ha side constraints!!
*****************************************************************************/
void ip_get_dy( int *iret)
{
  int i, j, cgitmax;
  double err, *beta1, *beta2, *dy2, *bdy2;
  
  *iret= 0;
  
  /* 1. factoritzem la part amb K blocs diagonals */
  ip_get_ASkAt(iret);
   if (*iret != 0) return;
  
  /* tractem cas trivial de problema uniarticle sense side constraints */
  if ((nat == 1) && (ncb == 0)) {
    dcopia(nb1,dy,bdy);
    ip_solve_ASkAt(dy);
    return;
  }
  
  /* 2. trobem y2: (D-C'*B^(-1)*C) y2 = (b2-C'*B^(-1)*b1) */ 
  beta1= beta;
  beta2= &beta[nb1];
  dy2= &(dy[nb1]);
  bdy2= &(bdy[nb1]);
  for(i=0;i<nb1;i++) beta1[i]= -bdy[i];
  ip_solve_ASkAt(beta1);
  ip_Ct_v(beta1,beta2);
  for(i=0;i<nb2;i++) beta2[i] += bdy2[i];
  
  /* tenim ja terme independent i resolem el sistema */
  ip_get_D();
  tipusprec= 0;
  if (tipusprec == 0) { /* gradient conjugat amb power series precond. */

/*
    if (sig_dig > 4) 
      m_pw_prec= 2;
    else
      m_pw_prec= 1;
*/    

/*    if (sig_dig > 3) 
      epscg= (double)1.0e-8;
    else
      epscg = MAX(epscg*(double)0.95,(double)1.0e-8);
*/

    epscg = MAX(epscg*(double)0.95,(double)1.0e-8);

    cgitmax= 2*nb2;
    /* for(i=0;i<nb2;i++) dy2[i]= (double)0.0; */
    cgit= dcg(nb2,ip_MCG_x,ip_H_x,dy2,beta2,rcg,zcg,pcg,cgitmax,epscg,&err);
  }
  if (cgit < 0) {
    *iret= 1;
    return;
  }
  else {
    *iret= 0;
    totcgit += cgit;
  }

  /* 3. trobem y1: B y1= (b1 - C y2) */
  ip_C_v(dy2,beta1);
  for(i=0;i<nb1;i++) dy[i] = bdy[i]-beta1[i];
  ip_solve_ASkAt(dy);
}

/*****************************************************************************
  Rutina per calcular per cada article k:  AS(k)A' i fer la seva factoritzacio.
  Omplenem lnz(.) i diag() per cada article. Un cop calculats realitza la
  seva factoritzacio usant rutines de SPARSPAK.
*****************************************************************************/
void ip_get_ASkAt(int *iret)
{
  int i,j,k;
  double *diagk, *lnzk, *sk, d;

  *iret= 0;
  for(k=0;k<nat;k++) {
    diagk= pdiag[k];
    lnzk= plnz[k];
    sk= psk[k];
    
    /* part diagonal */
    for(i=0;i<nnux;i++) {
      d= (double)0.0;
      for(j=ini_nus[i];j<ini_nus[i+1];j++) d += sk[iarc_incident[j]];
      diagk[i]= d;
    }
    
    /* part subdiagonal */
    for(i=0;i<maxlnz;i++) {
      d= (double)0.0;
      for(j= ini_arc[i];j<ini_arc[i+1];j++) d -= sk[iarc[j]];
      lnzk[i]= d;
    }
    
    /* factoritzacio */
    GSFCTINDEF1(&nnux,xlnz,lnzk,xnzsub,nzsub,diagk,w1spak,w2spak,
		 (double *)w3spak,iret); 
    if (*iret == 1) return;
  }
}

/*****************************************************************************
  Rutina per solucionar per tot article k: AS(k)A' x(k) = b(k)
  El vector d'entrada v es de dimensio nnux*nat, i conte per tota k el
  terme independent b(k). La solucio x(k) es deixa a sobre de b(k) (doncs
  es destrueix).
*****************************************************************************/
void ip_solve_ASkAt(double *v)
{
  int k,p;
  double *vk, *diagk, *lnzk;
  
  p=0;
  for(k=0;k<nat;k++) {
    diagk= pdiag[k];
    lnzk= plnz[k];
    vk= &(v[p]);
    GSSLV(&nnux,xlnz,lnzk,xnzsub,nzsub,diagk,vk);
    p += nnu;
  }
}

/*****************************************************************************
  Rutina per calcular C'v donat v vector real, i on la matriu C' es la que
  apareix en resoldre el sistema (ASA')dy=bb C'=(S1*A1'...  Sk*Ak') (nomes
  per als arcs de constriccions de cap. mutua actives).
  Retorna el resultat per ctv.
  v() te dimensio nb1=nnux*nat ; ctv() te dimensio nb2=m-nb1
  !!encara no hi ha side constraints!!
*****************************************************************************/
void ip_Ct_v(double *v, double *ctv)
{
  int i,k,pnu;
  double *vk, *sk, raux;
  
  for(i=0;i<nb2;i++) ctv[i]= (double)0.0; 
  if (nat > 1) {
    pnu= 0;
    for(k=0;k<nat;k++) {
      vk= &(v[pnu]);
      vk[nnux]= (double)0.0;       /* es considera posicio ficticia nnu */
      sk= psk[k];
      for(i=0;i<nxcmt;i++) {
	raux= (double)0.0; 
	raux += vk[mnk[iactcmt[i]]];  /* es considera posicio ficticia nnu */
	raux -= vk[mnl[iactcmt[i]]];  /* es considera posicio ficticia nnu */
	ctv[i] += sk[iactcmt[i]]*raux;
      }
      pnu += nnu;
    }
  }
}

/*****************************************************************************
  Rutina per calcular Cv donat v vector real, i on la matriu C es la que
  apareix en resoldre el sistema (ASA')dy=bb C'=(S1*A1'...  Sk*Ak') (nomes
  per als arcs de constriccions de cap. mutua actives).
  Retorna el resultat per cv.
  v() te dimensio nb2=m-nb1 ; cv() te dimensio nb1=nnux*nat
  !!encara no hi ha side constraints!!
*****************************************************************************/
void ip_C_v(double *v, double *cv)
{
  int i,k,pnu;
  double *cvk, *sk, raux;
  
  for(i=0;i<nb1;i++) cv[i]= (double)0.0; 
  if (nat > 1) {
    pnu= 0;
    for(k=0;k<nat;k++) {
      cvk= &(cv[pnu]);
      sk= psk[k];
      for(i=0;i<nxcmt;i++) {
	raux= v[i]*sk[iactcmt[i]];  
	cvk[mnk[iactcmt[i]]] += raux;  /* es considera posicio ficticia nnu */
	cvk[mnl[iactcmt[i]]] -= raux;  /* es considera posicio ficticia nnu */
      }
      cvk[nnux]= (double)0.0;  /* es considera posicio ficticia nnu */
      pnu += nnu;
    }
  }
}

/*****************************************************************************
  Rutina per obtenir D= Smc + \sum_k Sk per fer gradient conjugat
  Aquesta D es carrega a la variable d_gc i es pot usar com a 
  precondicionador diagonal
  !!no es contempla el cas de side constraints!!
*****************************************************************************/
void ip_get_D()
{
  int i,k;
  double *sk;

  for(i=0;i<nxcmt;i++) d_gc[i]= scmt[i];
  for(k=0;k<nat;k++) {
    sk= psk[k];
    for(i=0;i<nxcmt;i++) d_gc[i] += sk[iactcmt[i]];
  }
}
       
/*****************************************************************************
  Rutina per obtenir yy donat xx fent (D - C' B^(-1) C) xx = yy
  !!no es contempla el cas de side constraints!!
*****************************************************************************/
int ip_MCG_x(int nn, double *xx, double *yy)
{
  int i;
  
  /* usem les nat*nnux primeres posicions de beta() com espai de treball */
  ip_C_v(xx,beta);
  ip_solve_ASkAt(beta);
  ip_Ct_v(beta,yy);
  for(i=0;i<nb2;i++) yy[i]= d_gc[i]*xx[i]-yy[i];
  return(0);
}

/*****************************************************************************
  Rutina per obtenir zz donat rr resolent D zz = rr, quan s'usa com
  a precondicionador l'expansio en serie de potencies H=P^{-1}*Q, essent
  P= D  i  Q= C' B^(-1) C. L'ordre de l'expansio en serie de potencies
  ve donat per m_pw_prec;
  !!no es contempla el cas de side constraints!!
*****************************************************************************/
int ip_H_x(int nn, double *zz, double *rr, 
	   int (*atimes) (int, double *, double * ))
{
  int i,j;
  
  /* primer zz1 */
  for(j=0;j<nb2;j++) zz[j]= rr[j]/d_gc[j]; 
  
  /* despress zzi */
  for(i=1;i<m_pw_prec;i++) {
    /* usem les nat*nnux primeres posicions de beta() com espai de treball */
    ip_C_v(zz,beta);
    ip_solve_ASkAt(beta);
    ip_Ct_v(beta,zz);
    for(j=0;j<nb2;j++) zz[j]= (zz[j]+rr[j])/d_gc[j]; 
  }
  return(0);
}

/*****************************************************************************
  Rutina temporal per verificar que la direccio satisfa l'equacio de 
  Newton: Jx dk = -fk.
*****************************************************************************/
void ip_comprovar_direccio_es_correcta()
{
  int i,j,k,l;
  double *tmp1,*tmp2,r1,r2,*injk;

  tmp1= ALLOC(n+nat,double); /* es considera posicio ficticia nnu */
  if (!tmp1) exit(-1);
  tmp2= ALLOC(n+nat,double); /* es considera posicio ficticia nnu */
  if (!tmp2) exit(-1);
  
  /* comprovacio Zdx+Xdz = (mu - ZX) */
  printf("Equacio 1: Zdx+Xdz = (mu - ZX) \n");
  for(i=0;i<nar*nat;i++) {
    r1= z[i]*dx[i]+flux[i]*dz[i];
    r2= mu-z[i]*flux[i];
    printf("   residu (flux) %d= %f\n",i,r1-r2);
  }
  for(i=0;i<nxcmt;i++) {
    r1= zcmt[i]*dxcmt[i]+xcmt[i]*dzcmt[i];
    r2= mu-zcmt[i]*xcmt[i];
    printf("   residu (cmt) %d= %f\n",i,r1-r2);
  }
  for(i=0;i<ncb;i++) {
    r1= zcb[i]*dxcb[i]+xcb[i]*dzcb[i];
    r2= mu-zcb[i]*xcb[i];
    printf("   residu (cb) %d= %f\n",i,r1-r2);
  }
  
  /* comprovacio Adx= (b-Ax) */
  printf("Equacio 2: Adx= (b-Ax) \n");
  j= nnu*nat;
  ip_get_Ax(tmp1,&(tmp1[j]),&(tmp1[j+nxcmt]),dx,dxcmt,dxcb);
  ip_get_Ax(tmp2,&(tmp2[j]),&(tmp2[j+nxcmt]),flux,xcmt,xcb);
  for(k=0,l=0;k<nat;k++) {
    injk= pinj[k];
    for(i=0;i<nnux;i++) {
      tmp2[i] = injk[i]-tmp2[l];
      l++;
      printf("   residu (xarxa) %d= %f\n",i,tmp1[i]-tmp2[i]);  
    }
    l++; /* es considera posicio ficticia nnu */
  }
  for(i=0;i<nxcmt;i++) {
    tmp2[j+i] = cmt[iactcmt[i]]-tmp2[j+i];
    printf("   residu (cmt) %d= %f\n",j+i,tmp1[j+i]-tmp2[j+i]);
  }
  for(i=0;i<ncb;i++) {
    tmp2[j+nxcmt+i] = rhscb[i]-tmp2[j+nxcmt+i];
    printf("   residu (cb) %d= %f\n",j+nxcmt+i,tmp1[j+nxcmt+i]-tmp2[j+nxcmt+i]);
  }
  
  /* comprovacio -Wdx+Fdw = (mu - WF) */
  printf("Equacio 3: -Wdx+Fdw = (mu - WF) \n");
  for(i=0;i<nu;i++) {
    r1= -w[i]*dx[i]+f[i]*dw[i];
    r2= mu-w[i]*f[i];
    printf("   residu %d= %f\n",i,r1-r2);
  }

  /* comprovacio A(t)dy+dz-dw= (c-A(t)y-z+w) */
  printf("Equacio 4: A(t)dy+dz-dw= (c-A(t)y-z+w) \n");
  j= nar*nat;
  ip_get_Aty(tmp1,&(tmp1[j]),&(tmp1[j+nxcmt]),dy,dycmt,dycb);
  for(i=0;i<n;i++) tmp1[i] += dz[i];
  for(i=0;i<nu;i++) tmp1[i] -= dw[i];
  ip_get_Aty(tmp2,&(tmp2[j]),&(tmp2[j+nxcmt]),y,ycmt,ycb);
  for(i=0;i<nar*nat;i++) tmp2[i] = cost[i]-tmp2[i]-z[i];
  for(i=nar*nat;i<n;i++) tmp2[i] = -tmp2[i]-z[i];
  for(i=0;i<nu;i++) tmp2[i] += w[i];
  for(i=0;i<n;i++) {
    printf("   residu %d= %f\n",i,tmp1[i]-tmp2[i]);
  }

  free(tmp1);
  free(tmp2);
}
